/*=============================================================================
	UnXC_Game.h: Unreal game class.
=============================================================================*/

/*-----------------------------------------------------------------------------
	Unreal game engine.
-----------------------------------------------------------------------------*/
#if _MSC_VER
#pragma pack (push,4)
#endif

#ifndef XC_ENGINE_API
#define XC_ENGINE_API DLL_IMPORT
#endif

#ifndef NAMES_ONLY
#define AUTOGENERATE_NAME(name) extern XC_ENGINE_API FName XC_ENGINE_##name;
#define AUTOGENERATE_FUNCTION(cls,idx,name)
#endif

AUTOGENERATE_NAME(XC_Engine)
AUTOGENERATE_NAME(XC_GameEngine)
AUTOGENERATE_NAME(XC_ConnectionHandler)

#ifndef NAMES_ONLY

//Do remove debugging instructions
//#define CACUS_DEBUG

#include "XC_CoreObj.h"

//These variables should persist, so we can't really afford to let the garbage collector to mess this up

#ifndef DISABLE_COLLISION
	XC_ENGINE_API extern FCollisionCacus*	CollisionCacus;
#endif

/*
static void TestInverse()
{
	for ( FLOAT f=0 ; f<33000 ; f = f*1.4f + 0.005f )
	{
		FLOAT inv = _Reciprocal(f);
		FLOAT div = 1.f/f;
		debugf( NAME_Log, TEXT("Test inverse of %f: %f, %f, %f"), f, div, inv, div-inv);
	}
}
*/

#ifndef DISABLE_SCRIPT
class AXC_DummyPawn : public APawn
{
	friend class UXC_GameEngine;
	
	public:
	DECLARE_FUNCTION(NewMapName);			//539 - hack
	DECLARE_FUNCTION(NewPlayerCanSeeMe);		//532 - hack
	DECLARE_FUNCTION(NewDynArrayElement);		//10 - hack, version 10
	DECLARE_FUNCTION(NewDynArrayLength);		//640 - version 10
	DECLARE_FUNCTION(NewDynArrayInsert);		//641 - version 10
	DECLARE_FUNCTION(NewDynArrayRemove);		//642 - version 10
	DECLARE_FUNCTION(execAddToPackagesMap);		//1718 - version 11
	DECLARE_FUNCTION(execIsInPackageMap);		//1719 - version 13
	DECLARE_FUNCTION(PreLoginNative);		//3550 - internal
	DECLARE_FUNCTION(AdminLoginNative);		//3551 - internal
	DECLARE_FUNCTION(execCollidingActors);		//3552 - version 12
	DECLARE_FUNCTION(NewVirtualFunction);		//EX_VirtualFunction
	DECLARE_FUNCTION(PlayerSpeechNative);		//Direct binding - version 17
	DECLARE_FUNCTION(NewLocalVariable);			//Native replacement (bNetDirty) - version 17
	DECLARE_FUNCTION(NewInstanceVariable);		//Native replacement (bNetDirty) - version 17
	DECLARE_FUNCTION(NewDefaultVariable);		//Native replacement (bNetDirty) - version 17
	DECLARE_FUNCTION(NewLet);					//Native replacement (bNetDirty) - version 17
	DECLARE_FUNCTION(NewLetBool);				//Native replacement (bNetDirty) - version 17
	DECLARE_FUNCTION(NewContext);				//Native replacement (bNetDirty) - version 17
	DECLARE_FUNCTION(NewClassContext);			//Native replacement (bNetDirty) - version 17
	DECLARE_FUNCTION(NewStructMember);			//Native replacement (bNetDirty) - version 17

	//TODO:
	/*
	native(1715) final function Actor SpawnAct
	(
		Class<Actor>	SpawnClass,
		vector		SpawnLocation,
		optional rotator	SpawnRotation,
		optional name	ActorName,
		optional Actor	SpawnOwner,
		optional Pawn	SpawnInstigator,
		optional Actor	Template,
		optional bool	bMayColFail
	);
	*/
//    DECLARE_CLASS(APawn,AActor,0|CLASS_Abstract,XC_Engine)
};
#endif

//
// Editor entry point
//
class XC_ENGINE_API UXC_EditorLoader : public UObject
{
	public:
	DECLARE_CLASS(UXC_EditorLoader,UObject,0,XC_Engine)
};


//
// The Unreal game engine, subclassed.
//

struct FGameLevelHeader
{
	ULevel* GLevel;
	ULevel* GEntry;
	UPendingLevel* GPendingLevel;
	FURL			LastURL;
	TArray<FString> ServerActors;
	TArray<FString> ServerPackages;
};


#define XCGE_Defaults GetDefault<UXC_GameEngine>()
class XC_ENGINE_API UXC_GameEngine : public UGameEngine
{
	#include "UXC_GameEngine.h"
};



#endif

#ifndef NAMES_ONLY
#undef AUTOGENERATE_NAME
#undef AUTOGENERATE_FUNCTION
#endif NAMES_ONLY

#if _MSC_VER
#pragma pack (pop)
#endif

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
