//Find a file and attempt to compile whatever is in it

inline void NewLine( const TCHAR** Stream)
{
	while (**Stream == 0x000A || **Stream == 0x0085 || **Stream == 0x000D || **Stream == 0x000C || **Stream == 0x000B )
		(*Stream)++;
}

inline void Spacing( const TCHAR** Stream)
{
	while ( **Stream == ' ' || **Stream == 0x0009 )
		(*Stream)++;
}

inline INT ToNewLine( const TCHAR* Stream)
{
	INT Result = 0;
	while ( !(*Stream == 0x000A || *Stream == 0x0085 || *Stream == 0x000D || *Stream == 0x000C || *Stream == 0x000B) )
	{
		if ( !(*Stream) ) //End of file
			break;
		Stream++;
		Result++;
	}
	return Result;
}

inline INT ToEndComment( const TCHAR* Stream)
{
	INT Result = 0;
	while ( *Stream != '*' && *(Stream+1) != '/' )
	{
		if ( !(*Stream) ) //End of file
			break;
		Stream++;
		Result++;
	}
	return Result;
}

inline UBOOL NextStatement( const TCHAR** Stream)
{
	const TCHAR* OrgPos = *Stream;
	while ( 1 )
	{
		if ( **Stream == 0 ) //End of file
			return 0;
		NewLine( Stream);
		Spacing( Stream);
		if ( **Stream == '/' ) //A comment?
		{
			if ( *(*Stream+1) == '/' ) //Line comment
			{
				*Stream += ToNewLine( *Stream);
				continue;
			}
			else if ( *(*Stream+1) == '*' ) //Wrapped comment
			{
				*Stream += 2;
				*Stream += ToEndComment( *Stream);
				continue;
			}
		}
		if ( OrgPos == *Stream ) //No parsing has been done on this loop
			return 1;
		OrgPos = *Stream;
	}
	return 1;
}


inline INT FindOtherLimit( const TCHAR* Stream, TCHAR Delimiter)
{
	INT NextLine = ToNewLine( Stream);
	for ( int i=0 ; i<NextLine ; i++ )
	{
		if ( *(Stream+i) == Delimiter )
			return i;
	}
	return -1; //This will exit the compiler
}

inline UBOOL StopAt( const TCHAR** Stream, TCHAR tStopAt)
{
	INT i = FindOtherLimit( *Stream, tStopAt);
	if ( i < 0 )
	{
		debugf( NAME_Warning, TEXT("Expected a [%s] delimiter before %s..."), FString::Chr(tStopAt), FString::Printf(*Stream).Left(15) );
		return 0;
	}
	*Stream += (i + 1);
	return 1;
}

//These are mostly operators that can be put next to statements
inline UBOOL GenericSeparator( TCHAR Char, UBOOL bOperatorsOnly=0)
{
	if ( appIsAlnum(Char) || Char == '_' ) //Alphanumeric chars or lower dash
		return 0;
	if ( bOperatorsOnly )
		return (Char > 32 && Char < 127);
	return 1;
}

inline FString GenericCommand( const TCHAR** Stream, UBOOL bStatement=0)
{
	if ( bStatement && !NextStatement( Stream) )
	{
		debugf( NAME_Warning, TEXT("Unexpected EOF during command gathering") );
		return FString::Printf( TEXT("") );
	}
	INT i=0;
	const TCHAR* Cur = *Stream;
	while ( *Cur != 0 )
	{
		if ( GenericSeparator( *Cur ) ) //Any non alphanumeric character
		{
			if ( i == 0 )
				break;
			TArray<TCHAR> Res;
			Res.AddZeroed(i+1); //Zero char being the last char
			appMemcpy( &Res(0), *Stream, i*sizeof(TCHAR));
			*Stream += i;
			return *( (FString*) (DWORD) &Res);
		}
		Cur++;
		i++;
	}
	*Stream += i;
	return FString::Printf( *Stream - i );
}

inline FString GenericOperator( const TCHAR** Stream, UBOOL bStatement=0)
{
	if ( bStatement && !NextStatement( Stream) )
	{
		debugf( NAME_Warning, TEXT("Unexpected EOF during operator gathering") );
		return FString::Printf( TEXT("") );
	}
	if ( GenericSeparator(**Stream) ) //Alphanumeric operator
		return GenericCommand( Stream);
	INT i=0;
	const TCHAR* Cur = *Stream;
	while ( *Cur != 0 )
	{
		if ( !GenericSeparator( *Cur, true) ) //Operators only
		{
			if ( i == 0 )
				break;
			TArray<TCHAR> Res;
			Res.AddZeroed(i+1); //Zero char being the last char
			appMemcpy( &Res(0), *Stream, i*sizeof(TCHAR));
			*Stream += i;
			return *( (FString*) (DWORD) &Res);
		}
		Cur++;
		i++;
	}
	*Stream += i;
	return FString::Printf( *Stream - i );
}

inline UBOOL IsFunction( const TCHAR* Stream) {	return ( *Stream == '(' ); }

void UXC_GameEngine::CompileText( FString FileName)
{
	guard(UXC_GameEngine::CompileText);

	debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("CompileText: Compiling %s..."), *FileName );

	FString Code;
	if ( !appLoadFileToString( Code, *FileName) )
	{
		debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("CompileText: file %s not found"), *FileName );
		return;
	}

	const TCHAR* Stream = *Code;

	//Find opening text
	while ( NextStatement( &Stream) )
	{
		FString Command = GenericCommand( &Stream);
		if ( Command == TEXT("edit_function") )
		{
			if ( !NextStatement( &Stream) )
			{
				debugf( NAME_Warning, TEXT("Failed to compile, EOF after [edit_function] statement") );
				return;
			}
			if ( !IsFunction(Stream++) )
			{
				debugf( NAME_Warning, TEXT("Failed to compile, [edit_function] statement doesn't have a parameter block") );
				return;
			}
			UPackage* FuncPkg = NULL;
			UClass* FuncClass = NULL;
			UState* FuncState = NULL;
			Command = GenericCommand( &Stream, 1);
			if ( Command == TEXT("Package") )
			{
				Command = GenericOperator( &Stream, 1);
				if ( Command != TEXT("=") )
				{
					debugf( NAME_Warning, TEXT("Failed to compile, No [=] operator after [Package] keyword") );
					return;
				}
				Command = GenericCommand( &Stream, 1);
				FuncPkg = (UPackage*) UObject::StaticFindObject( UPackage::StaticClass(), NULL, *Command, 1);
				if ( !FuncPkg )
					debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("CompileText: Package %s not found, class will be searched without Outer package"), *Command);
			}
			else if ( Command == TEXT("Class") )
			{
				Command = GenericOperator( &Stream, 1);
				if ( Command != TEXT("=") )
				{
					debugf( NAME_Warning, TEXT("Failed to compile, No [=] operator after [Class] keyword") );
					return;
				}
				Command = GenericCommand( &Stream, 1);
				FuncClass = (UClass*) UObject::StaticFindObject( UClass::StaticClass(), FuncPkg, *Command, 1);
				if ( !FuncClass )
				{
					debugf( NAME_Warning, TEXT("Failed to compile, class %s not found"), *Command);
					return;
				}
			}
			else if ( Command == TEXT("State") )
			{
				if ( FuncClass )
				{	Command = GenericOperator( &Stream, 1);
					if ( Command != TEXT("=") )
					{
						debugf( NAME_Warning, TEXT("Failed to compile, No [=] operator after [State] keyword") );
						return;
					}
					Command = GenericCommand( &Stream, 1);
					FuncState = (UState*) UObject::StaticFindObject( UState::StaticClass(), FuncClass, *Command, 1); //Should use a field iterator instead
					if ( !FuncState )
						debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("CompileText: State %s not found, global function will be searched instead"), *Command);
				}
				else
					debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("CompileText: Search for class first!, ignoring [State] parameter") );
			}
		}
	}


/*	UFunction* PreLogin = FindBaseFunction( AGameInfo::StaticClass(), TEXT("PreLogin"));
	if ( !PreLogin ) //Script not serialized?
		return;

	GNatives[3550] = &UObject::execUndefined;
	if ( !OldPreLogin.Num() )
	{
		if ( !bSet )
			return;
		OldPreLogin = PreLogin->Script;
	}
	
	if ( !bSet )
	{
		check( OldPreLogin.Num() );
		PreLogin->Script = OldPreLogin;
		debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("Reverting GameInfo.PreLogin hook") );
		return;
	}
	else if ( OldPreLogin.Num() != PreLogin->Script.Num() )
		return;

	//Initialize remaining vars
	INT Found = 0;
	UProperty* Options = FindInstanceVariable( PreLogin, TEXT("Options"), &Found);
	UProperty* Address = FindInstanceVariable( PreLogin, TEXT("Address"), &Found);
	UProperty* Error = FindInstanceVariable( PreLogin, TEXT("Error"), &Found);
	UProperty* FailCode = FindInstanceVariable( PreLogin, TEXT("FailCode"), &Found);
	check( Found == 4 );
	
	PreLogin->Script.InsertZeroed( PreLogin->Script.Num()-2, 23 ); //Add 23 bytes right before the return
	BYTE* Block = &PreLogin->Script( PreLogin->Script.Num() - 25);

	GNatives[3550] = (Native)&UXC_GameEngine::PreLoginNative;
	HighNative( Block, 3550);				Block += 2;
	ParseLocalVariable( Block, Options);	Block += 5;
	ParseLocalVariable( Block, Address);	Block += 5;
	ParseLocalVariable( Block, Error);		Block += 5;
	ParseLocalVariable( Block, FailCode);	Block += 5;
	*Block++ = 0x16; //EX_EndFunctionParms
	check( *Block == 0x04); //Beginning of return

	debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("GameInfo.PreLogin hook success"));*/
	unguard;
}
