
#include "XC_Engine.h"
#include "UnNet.h"
#include "UnXC_Travel.h"
#include "UnXC_Col.h"
#include "UnXC_TimeManager.h"
#include "UnXC_ConnHandler.h"

//XC_CORE!!
#include "XC_CoreClasses.h"
#include "XC_CoreGlobals.h"
#include "XC_Download.h"

IMPLEMENT_CLASS(UXC_GameEngine);
IMPLEMENT_CLASS(UXC_EditorLoader);
IMPLEMENT_CLASS(AXC_ConnectionHandler);
IMPLEMENT_PACKAGE(XC_Engine);


#define NAMES_ONLY
#define AUTOGENERATE_NAME(name) XC_ENGINE_API FName XC_ENGINE_##name;
#define AUTOGENERATE_FUNCTION(cls,idx,name) IMPLEMENT_FUNCTION(cls,idx,name)
#include "UnXC_Game.h"
#undef AUTOGENERATE_FUNCTION
#undef AUTOGENERATE_NAME
#undef NAMES_ONLY
void RegisterNames()
{
	static INT Registered=0;
	if(!Registered++)
	{
		#define NAMES_ONLY
		#define AUTOGENERATE_NAME(name) extern XC_ENGINE_API FName XC_ENGINE_##name; XC_ENGINE_##name=FName(TEXT(#name),FNAME_Intrinsic);
		#define AUTOGENERATE_FUNCTION(cls,idx,name)
		#include "UnXC_Game.h"
		#undef DECLARE_NAME
		#undef NAMES_ONLY
	}
}

struct FArrayHack
{
	void* Data;
	INT	  ArrayNum;
	INT	  ArrayMax;
	inline INT Num()
	{
		return ArrayNum;
	}
};

#ifndef DISABLE_SCRIPT
struct STATIC_ARRAY
{
	const INT Size;
	const INT DataSize;
	INT Ceiling;

	STATIC_ARRAY( DWORD NewSize=0, DWORD NewDataSize=4)
	:	Size(NewSize)
	,	DataSize(NewDataSize)
	,	Ceiling(0)
	{};
	inline UBOOL IsCleared() { return !Ceiling; }
	inline void Clear() { Ceiling=0; }
};

#define _STATIC_ARRAY(a_name,size,type) \
		struct SA_##a_name : public STATIC_ARRAY \
		{ \
			type Data[size]; \
			SA_##a_name() : STATIC_ARRAY(size,sizeof(type)) { appMemzero(&Data, sizeof(type) * size); } \
			UBOOL Setup( type* Other) { appMemcpy_amd( &Data, Other, size * sizeof(type) ); return 1; } \
			UBOOL Setup( TArray<type>* Other) { if (Other->Num() > Size) return 0; appMemcpy(&Data, &((*Other)(0)), sizeof(type) * Other->Num()); Ceiling = Other->Num(); return 1; } \
			void Restore( type* Other) { appMemcpy_amd( Other, &Data, size * sizeof(type) ); } \
			void Restore( TArray<type>* Other) { SafeEmptyR(Other); Other->Add(Ceiling); appMemcpy_amd( &((*Other)(0)), &Data, sizeof(type) * Other->Num()); } \
			inline UBOOL MismatchSize( TArray<type>* Other) { return Ceiling != Other->Num(); } \
		}; \
		static SA_##a_name a_name = SA_##a_name();

//Keeping these 4-byte aligned
_STATIC_ARRAY(OldGetWeapon,280,BYTE);
_STATIC_ARRAY(OldPreLogin,252,BYTE);
_STATIC_ARRAY(OldAdminLogin,32,BYTE);
_STATIC_ARRAY(OldSetEnemy,1360,BYTE);
_STATIC_ARRAY(OldRocketTick,344,BYTE);
_STATIC_ARRAY(OldSetOrders,1200,BYTE);
_STATIC_ARRAY(OldShowPath,132,BYTE);
_STATIC_ARRAY(OldShowInventory,252,BYTE);
_STATIC_ARRAY(OldForceFire,2,BYTE);
_STATIC_ARRAY(OldForceAltFire,2,BYTE);

//Individual functions
#include "_GetWeapon_func.h"
#include "_PreLogin_func.h"
#include "_AdminLogin_func.h"
#include "_ushare_SetEnemy_func.h"
#include "_uteightball_Tick_func.h"
#include "_botpack_SetOrders_func.h"
#include "_ShowPath_func.h"
#include "_ShowInventory_func.h"
#include "_ForceFire_func.h"
#include "_ForceAltFire_func.h"
#include "Functions/_PlayerSpeech_func.h"
#include "Functions/_UWindowListSort_func.h"
#include "Functions/_PlayerPawn_funcs.h"
#endif

static UBOOL		bGCEnded = 0;
static AActor*		GPropActor = NULL;

#include "XCGE_Private.h"

void UXC_GameEngine::PaintProgress()
{
	guard(PaintProgress);

	FVector LoadFog(0.20,.02,.02);
	FVector LoadScale(.2,.2,.2);
	UViewport* Viewport=Client->Viewports(0);
	Exchange(Viewport->Actor->FlashFog,LoadFog);
	Exchange(Viewport->Actor->FlashScale,LoadScale);
	Draw( Viewport );
	Exchange(Viewport->Actor->FlashFog,LoadFog);
	Exchange(Viewport->Actor->FlashScale,LoadScale);

	unguard;
}

//
// Serializer.
// We cannot serialize in v451 linux, let UProperty handle it
void UXC_GameEngine::Serialize( FArchive& Ar )
{
	guard(UXC_GameEngine::Serialize);

	for ( int i=0 ; i<12 ; i++ )
		if ( PreLoginHooks[i] && PreLoginHooks[i]->bDeleteMe )
			PreLoginHooks[i] = NULL;
	bGCEnded = true;

	Super::Serialize( Ar );
	unguard;
}


UXC_GameEngine::UXC_GameEngine()
{
	LoadConfig();
}

//
// Class creator.
//
#ifndef DISABLE_STATICS
void UXC_GameEngine::StaticConstructor()
{
	guard(UXC_GameEngine::StaticConstructor);

	UClass* TheClass = GetClass();
	UXC_GameEngine* DefaultObject = (UXC_GameEngine*) &TheClass->Defaults(0);

	DefaultObject->Version = 17;
	DefaultObject->CoreVersion = 3;

	for( TFieldIterator<UProperty> FIt( UGameEngine::StaticClass() ); FIt; ++FIt )
		if( (appStricmp( FIt->GetName(), TEXT("MinClientVersion") ) == 0) && (FIt->Offset == 104) )
		{
			if ( sizeof(class UEngine) == 108 )
			{
//				debugf( TEXT("USING V451 HEADERS"));
			}
			else
				DefaultObject->b451Setup = true;
			break;
		}

	DWORD NativeConst = CPF_Native|CPF_Edit|CPF_EditConst|CPF_Const;
	new( TheClass,TEXT("XC_Version"), RF_Public) UIntProperty( CPP_PROPERTY(Version), TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("XC_Core_Version"), RF_Public) UIntProperty( CPP_PROPERTY(CoreVersion), TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("AdminLoginHook"), RF_Public) UObjectProperty( CPP_PROPERTY(AdminLoginHook), TEXT("XC_GameEngine"), CPF_Edit, AActor::StaticClass() );

	UObjectProperty* PreLoginHooksArray = new( TheClass,TEXT("PreLoginHooks"), RF_Public) UObjectProperty( CPP_PROPERTY(PreLoginHooks), TEXT("XC_GameEngine"), CPF_Edit, AActor::StaticClass() );
	PreLoginHooksArray->ElementSize = 4;
	PreLoginHooksArray->ArrayDim = 12;

	new( TheClass,TEXT("bForceLevelHook"), RF_Public) UBoolProperty(CPP_PROPERTY(DummyData_1) - 4, TEXT("XC_GameEngine"), CPF_Native | CPF_Edit | CPF_Config);
	new( TheClass,TEXT("bScriptDebug"), RF_Public) UBoolProperty(CPP_PROPERTY(DummyData_1) - 4, TEXT("XC_GameEngine"), CPF_Native | CPF_Edit | CPF_Config);
	new( TheClass,TEXT("bAutoCompressLZMA"), RF_Public) UBoolProperty(CPP_PROPERTY(DummyData_1) - 4, TEXT("XC_GameEngine"), CPF_Native | CPF_Edit | CPF_Config);
	new( TheClass,TEXT("bUseLevelHook"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	DefaultObject->bUseLevelHook = 1;
	new( TheClass,TEXT("bUseSetEnemyHook"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	DefaultObject->bUseSetEnemyHook = 1;
	new( TheClass,TEXT("bUseNewRelevancy"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	DefaultObject->bUseNewRelevancy = 1;
	new( TheClass,TEXT("bEnableDebugLogs"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	new( TheClass,TEXT("bFasterUpload"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	DefaultObject->bFasterUpload = 1; //Defaults to 1
	new( TheClass,TEXT("bCollisionHashHook"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	DefaultObject->bCollisionHashHook = 1; //Defaults to 1
	new( TheClass,TEXT("bCacheConvertAtJoin"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	new( TheClass,TEXT("bAutoTravelManager"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	new( TheClass,TEXT("bSortMaplistGlobal"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	new( TheClass,TEXT("bSortMaplistByFolder"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	new( TheClass,TEXT("bDisableBrushTracker"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	new( TheClass,TEXT("bDisableTimingFix"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_1)-4, TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );

	//UBool properties on different bitfields must not be defined together, or else everything's gon' get fucked up (bitmaks become consequent when they shouldn't)
	UArrayProperty *TrackerMapArray = new( TheClass,TEXT("NoBrushTrackerFix"), RF_Public) UArrayProperty ( CPP_PROPERTY(NoBrushTrackerFix), TEXT("XC_GameEngine"), CPF_Native|CPF_Edit|CPF_Config );
	TrackerMapArray->Inner = new (TrackerMapArray, TEXT("NoBrushTrackerFix"), RF_Public) UStrProperty;
	TrackerMapArray->Inner->PropertyFlags = CPF_Edit|CPF_Config;
	TrackerMapArray->Inner->ElementSize = 12;

	new( TheClass,TEXT("bHTTP_DLMGR"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("b440Net"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("b451Setup"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("bCanEditPackageMap"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("bSetOrders"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("bRocketTick"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("bSetEnemy"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("bHackedEngine"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );
	new( TheClass,TEXT("bHackingTracker"), RF_Public) UBoolProperty( CPP_PROPERTY(DummyData_2)-4, TEXT("XC_GameEngine"), NativeConst );

	//UGameEngine
	new( UGameEngine::StaticClass(), TEXT("Level"), RF_Public) UObjectProperty( EC_CppProperty, 104 + 4 * DefaultObject->b451Setup, TEXT("GameEngine"), NativeConst , ULevel::StaticClass() );
	new( UGameEngine::StaticClass(), TEXT("Entry"), RF_Public) UObjectProperty( EC_CppProperty, 108 + 4 * DefaultObject->b451Setup, TEXT("GameEngine"), NativeConst, ULevel::StaticClass() );
	new( UGameEngine::StaticClass(), TEXT("PendingLevel"), RF_Public) UObjectProperty( EC_CppProperty, 112 + 4 * DefaultObject->b451Setup, TEXT("GameEngine"), NativeConst, ULevelBase::StaticClass() );

	//UDemoRecDriver
	new( UDemoRecDriver::StaticClass(), TEXT("DemoFilename"), RF_Public) UStrProperty( EC_CppProperty, UNetDriver::StaticClass()->GetPropertiesSize() , TEXT("DemoRecDriver"), NativeConst ); //SHOULD BE 152

	//UNetDriver
	UArrayProperty *UNetCons = new( UNetDriver::StaticClass(), TEXT("ClientConnections"), RF_Public) UArrayProperty( EC_CppProperty, 44, TEXT("NetDriver"), NativeConst);
	UNetCons->Inner = new (UNetCons, TEXT("ObjectProperty0"), RF_Public) UObjectProperty;
	((UObjectProperty*) UNetCons->Inner)->PropertyClass = UNetConnection::StaticClass();
	UNetCons->Inner->PropertyFlags = CPF_Edit|CPF_EditConst|CPF_Const;
	UNetCons->Inner->ElementSize = 4;
	new( UNetDriver::StaticClass(), TEXT("ServerConnection"), RF_Public) UObjectProperty( EC_CppProperty, 56, TEXT("NetDriver"), NativeConst, UNetConnection::StaticClass() );

	//ULevel
	new( ULevelBase::StaticClass(), TEXT("NetDriver"), RF_Public) UObjectProperty( EC_CppProperty, 60, TEXT("LevelBase"), NativeConst, UNetDriver::StaticClass() );
	new( ULevelBase::StaticClass(), TEXT("Engine"), RF_Public) UObjectProperty( EC_CppProperty, 64, TEXT("LevelBase"), NativeConst, UEngine::StaticClass() );
	//Need URL struct
	new( ULevelBase::StaticClass(), TEXT("URL_Protocol"), RF_Public) UStrProperty( EC_CppProperty, 68 , TEXT("LevelBase"), NativeConst );
	new( ULevelBase::StaticClass(), TEXT("URL_Host"), RF_Public) UStrProperty( EC_CppProperty, 80 , TEXT("LevelBase"), NativeConst );
	new( ULevelBase::StaticClass(), TEXT("URL_Port"), RF_Public) UIntProperty( EC_CppProperty, 92 , TEXT("LevelBase"), NativeConst );
	new( ULevelBase::StaticClass(), TEXT("URL_Map"), RF_Public) UStrProperty( EC_CppProperty, 96 , TEXT("LevelBase"), NativeConst );
	UArrayProperty *URLOptions = new( ULevelBase::StaticClass(),TEXT("URL_Options"), RF_Public) UArrayProperty ( EC_CppProperty, 108, TEXT("LevelBase"), NativeConst );
	URLOptions->Inner = new (URLOptions, TEXT("StringProperty0"),RF_Public) UStrProperty;
	URLOptions->Inner->PropertyFlags = NativeConst;
	URLOptions->Inner->ElementSize = 12;
	new( ULevelBase::StaticClass(), TEXT("URL_Portal"), RF_Public) UStrProperty( EC_CppProperty, 120 , TEXT("LevelBase"), NativeConst );
	//URL end
	new( ULevelBase::StaticClass(), TEXT("DemoRecDriver"), RF_Public) UObjectProperty( EC_CppProperty, 136, TEXT("LevelBase"), NativeConst , UNetDriver::StaticClass() );
	new( ULevelBase::StaticClass(), TEXT("ActorListSize"), RF_Public) UIntProperty( EC_CppProperty, 48 , TEXT("LevelBase"), NativeConst );

	new( ULevel::StaticClass(), TEXT("iFirstNetRelevantActor"), RF_Public) UIntProperty( EC_CppProperty, 276 , TEXT("Level"), NativeConst );
	new( ULevel::StaticClass(), TEXT("iFirstDynamicActor"), RF_Public) UIntProperty( EC_CppProperty, 272 , TEXT("Level"), NativeConst );

	//AActor - unrealscript serialized, cannot be added here
//	new( AActor::StaticClass(),TEXT("bNotRelevantToOwner"), RF_Public) UBoolProperty( EC_CppProperty, 512, TEXT("Networking"), CPF_Native|CPF_Edit );

	
	unguard;
}
#endif

UBOOL UXC_GameEngine::Exec( const TCHAR* Cmd, FOutputDevice& Ar)
{
	guard(UXC_GameEngine::Exec);
	const TCHAR* Str = Cmd;

	if ( ParseCommand(&Cmd,TEXT("DumpObject")) )
	{
		FName ObjectName;
		FName OuterName;
		UBOOL UseOuter = 0;
		UObject* Found = NULL;

		if ( Parse( Cmd, TEXT("Outer="), OuterName ) )
			UseOuter = true;
		if( Parse( Cmd, TEXT("Name="), ObjectName ) )
		{
			// look for object by name
			for( TObjectIterator<UObject> It; It; ++It )
			{	if( It->GetName() == *ObjectName )
				{
					if ( UseOuter && It->GetOuter() && It->GetOuter()->GetName() != *OuterName )
						continue;
					Found = *It;
					break;
				}
			}
		}
		if( Found )
		{
			FArchive* Src = GFileManager->CreateFileWriter( *ObjectName, (true?0:FILEWRITE_NoReplaceExisting) | (true?FILEWRITE_EvenIfReadOnly:0) );
			Src->Serialize( Found, Found->GetClass()->GetPropertiesSize() );
			Src->Close();
			delete Src;

			if ( Found->GetClass() == UFunction::StaticClass() )
			{
				FArchive* uSrc = GFileManager->CreateFileWriter( TEXT("FUNCTIONDATA.bin"), (true?0:FILEWRITE_NoReplaceExisting) | (true?FILEWRITE_EvenIfReadOnly:0) );
				FArrayHack* ScriptArray = (FArrayHack*) &((UFunction*)Found)->Script;
				uSrc->Serialize( ScriptArray->Data, ScriptArray->Num() );
//				*uSrc << ((UFunction*) Found)->Script; //No byte array mangling here for now
				uSrc->Close();
			}
//			FString* TestStr = (FString*) ( (DWORD)Found + 0x0268 );
//			debugf( NAME_Log, TEXT("[XC_ENGINE] ACTOR DUMPED, %s, %i"), *TestStr, Found->GetClass()->GetPropertiesSize() );
		}
	}
	else if ( ParseCommand(&Cmd,TEXT("XC_Engine")) || ParseCommand(&Cmd,TEXT("XC_Version")) )
	{
		Ar.Logf( NAME_Log, TEXT("Running XC_Engine version %i, made by Higor"), Version );
		Ar.Log( *FString::Printf( TEXT("%i"), Version) );
	}
	else if ( ParseCommand(&Cmd,TEXT("OctreeDebug")) )
	{
		FName ActorName;
		if ( Parse( Cmd, TEXT("Name="), ActorName ) )
		{
			FCacusActorInfo* AInfo = CollisionCacus->FindActor( ActorName);
			if ( AInfo && AInfo->Actor)
			{
				const FCollisionOctree* Node = CollisionCacus->GridData->FullList( AInfo->Actor->CollisionTag);
				Ar.Logf( NAME_Log, TEXT("Actor %s Octree info:"), *ActorName );
				Ar.Logf( NAME_Log, TEXT("== Stored in node %i, with depth %i and with %i children"), AInfo->Actor->CollisionTag, Node->CurDepth, Node->ChildCount);
				Ar.Logf( NAME_Log, TEXT("== Node bounds (%i,%i,%i) to (%i,%i,%i)"), appRound(Node->BoundingBox.Min.X), appRound(Node->BoundingBox.Min.Y), appRound(Node->BoundingBox.Min.Z),
																			appRound(Node->BoundingBox.Max.X), appRound(Node->BoundingBox.Max.Y), appRound(Node->BoundingBox.Max.Z));
				Ar.Logf( NAME_Log, TEXT("== Node grid position (%i,%i,%i), with cubic grid of size %i"), Node->XPos, Node->YPos, Node->ZPos, CollisionCacus->DepthGrid[Node->CurDepth].GridSize);
				if ( AInfo->TopDepth > Node->CurDepth )
					Ar.Logf( NAME_Log, TEXT("== Actor being temporarily held in this node until more arrive (Target Depth: %i)"), AInfo->TopDepth);
				Ar.Logf( NAME_Log, TEXT("== Actor Bounds (%i,%i,%i) to (%i,%i,%i)"), appRound(AInfo->Box.Min.X), appRound(AInfo->Box.Min.Y), appRound(AInfo->Box.Min.Z),
																			appRound(AInfo->Box.Max.X), appRound(AInfo->Box.Max.Y), appRound(AInfo->Box.Max.Z));
				Ar.Logf( NAME_Log, TEXT("== Cached Location (%i,%i,%i)"), appRound(AInfo->Location.X), appRound(AInfo->Location.Y), appRound(AInfo->Location.Z));
			}
		}
	}
	else if ( ParseCommand(&Cmd,TEXT("LogFields")) )
	{
		FName ClassName;
		UClass* Found = NULL;

		if( Parse( Cmd, TEXT("Name="), ClassName ) )
		{
			// look for object by name
			for( TObjectIterator<UClass> It; It; ++It )
				if( It->GetName() == *ClassName )
				{
					Found = *It;
					break;
				}
		}
		if( Found )
		{
			debugf( NAME_XC_Engine, TEXT("Dumping property list for %s"), Found->GetName() );
			for( TFieldIterator<UProperty> It(Found); It; ++It )
			{
				if ( It->GetClass() == UBoolProperty::StaticClass() )
					debugf( NAME_XC_Engine, TEXT("Property %s: Offset=%i, Flags=%i, Bitmask=%i, Size=%i"), It->GetName(), It->Offset, It->PropertyFlags, ((UBoolProperty*)(*It))->BitMask, It->ElementSize);
				else
					debugf(NAME_XC_Engine, TEXT("Property %s: Offset=%i, Flags=%i, Array=%i, Size=%i"), It->GetName(), It->Offset, It->PropertyFlags, It->ArrayDim, It->ElementSize);
			}
		}
	}
	else if ( ParseCommand(&Cmd,TEXT("LogClassSizes")) )
	{
		FName OuterName;

		if( Parse( Cmd, TEXT("Outer="), OuterName ) )
		{
			debugf(NAME_Log, TEXT("[XC_Engine] Logging class sizes for classes residing in %s"), *OuterName);
			// look for object by name
			for( TObjectIterator<UClass> It; It; ++It )
				if( It->GetOuter() && It->GetOuter()->GetName() == *OuterName )
					debugf(NAME_XC_Engine, TEXT("%s => %i"), It->GetName(), It->GetPropertiesSize() );
		}
		else
		{
			debugf(NAME_XC_Engine, TEXT("Logging all class sizes"));
			// look for object by name
			for( TObjectIterator<UClass> It; It; ++It )
				if( It->GetOuter()  )
					debugf(NAME_XC_Engine, TEXT("[%s] %s => %i"), It->GetOuter()->GetName(), It->GetName(), It->GetPropertiesSize() );
		}
	}
#ifndef __LINUX_X86__
	else if ( ParseCommand(&Cmd,TEXT("EditObject")) )
	{
		FName ObjectName;
		UObject* Found = NULL;

		if( Parse( Cmd, TEXT("Name="), ObjectName ) )
		{
			FString aStr;
			INT Skip = 0;
			if ( Parse( Cmd, TEXT("Skip="), aStr) )
				Skip = appAtoi( *aStr );
			// look for object by name
			for( TObjectIterator<UObject> It; It; ++It )
			{
				//Do not bring up classes
				if( (It->GetName() == *ObjectName) && !It->IsA( UClass::StaticClass()) )
				{
					Found = *It;
					if ( Skip-- == 0 )
						break;
				}
			}
		}
		if( Found )
		{
			WObjectProperties* P = new WObjectProperties( TEXT("EditObject"), 0, TEXT(""), NULL, 1 );;
			P->OpenWindow( (HWND)Client->Viewports(0)->GetWindow() );
			P->Root.SetObjects( (UObject**)&Found, 1 );
			P->Show(1);
		}
	}
#endif
	else if ( ParseCommand(&Cmd,TEXT("ToggleTimingFix")) || ParseCommand(&Cmd,TEXT("TimingFix")) )
	{
		bDisableTimingFix = !bDisableTimingFix;
		SaveConfig();
		Ar.Logf( NAME_XC_Engine, TEXT("Timing fix %s"), bDisableTimingFix ? TEXT("disabled") : TEXT("enabled") );
	}
	else if ( ParseCommand(&Cmd,TEXT("ToggleDebugLogs")) || ParseCommand(&Cmd,TEXT("DebugLogs")) )
	{
		bEnableDebugLogs = !bEnableDebugLogs;
		SaveConfig();
		Ar.Logf( NAME_XC_Engine, TEXT("Debug Logs %s"), bEnableDebugLogs ? TEXT("enabled") : TEXT("disabled") );
	}
	else if ( ParseCommand(&Cmd,TEXT("ToggleRelevancy")) || ParseCommand(&Cmd,TEXT("ToggleRelevant")) )
	{
		bUseNewRelevancy = !bUseNewRelevancy;
		SaveConfig();
		Ar.Logf( NAME_XC_Engine, TEXT("Advanced relevancy %s"), bUseNewRelevancy ? TEXT("enabled") : TEXT("disabled") );
	}
#ifdef CACUS_DEBUG
	else if ( ParseCommand(&Cmd,TEXT("ActorSlots")) )
	{
		APlayerPawn* PlayerActor = NULL;
		if ( Client && Client->Viewports.Num() && Client->Viewports(0) && Client->Viewports(0)->Actor )
			PlayerActor = Client->Viewports(0)->Actor;
		int i = 0;
		while ( i < PlayerActor->XLevel->Actors.Num() )
		{
			AActor* TestActor = PlayerActor->XLevel->Actors(i++);
			if ( !TestActor )
				debugf( NAME_XC_Engine, TEXT("==Free Actor slot at %i"), i);
			else if ( TestActor->bDeleteMe )
				debugf( NAME_XC_Engine, TEXT("==Deleted Actor slot at %i"), i);
			else if ( TestActor == PlayerActor )
				break;
		}
	}
#endif
	else if ( !b451Setup && ParseCommand(&Str,TEXT("SET")) )
	{
		while ( *Str == ' ' ) Str++;
		if ( appStrnicmp( Str, TEXT("ini:"), 4 ) == 0 )
			return Super::Exec( Cmd, Ar);
		PerformSetCommand( Str, Ar, 1);
	}
	else if ( !b451Setup && ParseCommand(&Str,TEXT("GET")) ) //Add v451 functionality in v436 + better stuff
	{
		while ( *Str == ' ' ) Str++;
		if ( appStrnicmp( Str, TEXT("ini:"), 4 ) == 0 )
			return Super::Exec( Cmd, Ar);
		// Get a class default variable.
		TCHAR ClassName[128], PropertyName[128], ArrayIdx[32];
		UClass* Class;
		UProperty* Property;
		if	(	ParseToken( Str, ClassName, ARRAY_COUNT(ClassName), 1 )
		&&	(Class=FindObject<UClass>( ANY_PACKAGE, ClassName))!=NULL )
		{
			if	(	ParseToken( Str, PropertyName, ARRAY_COUNT(PropertyName), 1 )
			&&	(Property=FindField<UProperty>( Class, PropertyName))!=NULL )
			{
				INT Idx = 0;
				if ( ParseToken( Str, ArrayIdx, ARRAY_COUNT(ArrayIdx), 1) )
					Idx = Clamp(appAtoi( ArrayIdx), 0, Property->ArrayDim - 1);
				TCHAR Temp[256];
				if( Class->GetPropertiesSize() )
					Property->ExportText( Idx, Temp, (BYTE*)(Class->GetDefaultObject()), (BYTE*)(Class->GetDefaultObject()), PPF_Localized );
				Ar.Log( Temp );
			}
			else Ar.Logf( NAME_ExecWarning, TEXT("Unrecognized property %s"), PropertyName );
		}
		else Ar.Logf( NAME_ExecWarning, TEXT("Unrecognized class %s"), ClassName );
	}
	else if ( ParseCommand(&Str,TEXT("PreLoginHook")) )
	{
		TCHAR ActorName[65];
		if ( ParseToken( Str, ActorName, 65, 1 ) )
		{
			AActor* Found = FindNamedActor( GetActiveLevel(this), ActorName);
			if ( !Found )
				Ar.Logf( TEXT("Actor %s not found in level"), ActorName );
			else
			{
				UFunction* PreLoginHook = FindBaseFunction( Found->GetClass(), TEXT("PreLoginHook") );
				if ( !PreLoginHook )
					Ar.Logf( TEXT("Class %s doesn't have a PreLoginHook function"), Found->GetClass()->GetName() );
				else
				{
					if ( PreLoginHook->ParmsSize != 48 )
						debugf( NAME_Warning, TEXT("%s.PreLoginHook has incorrect parameter block size: %i, should be 48"), ActorName, PreLoginHook->ParmsSize );
					else
					{
						int SuccessCount = 0;
						UProperty* Parm = (UProperty*)PreLoginHook->Children;
						INT i;
						for ( i=0 ; i<4 ; i++ )
						{
							if ( Parm->GetClass() == UStrProperty::StaticClass() )
								SuccessCount++;
							else debugf( NAME_Warning, TEXT("%s.PreLoginHook has incorrect parameter %i type, should be STRING"), ActorName, i+1 );
							if ( i > 1 )
							{
								if ( Parm->PropertyFlags & CPF_OutParm )
									SuccessCount++;
								else debugf( NAME_Warning, TEXT("%s.PreLoginHook parameter %i must have OUT flag!"), ActorName, i+1 );
							}
							Parm=(UProperty*)Parm->Next;
						}
						if ( SuccessCount == 6 )
						{
							for ( i=0 ; i<12 ; i++ )
								if ( !PreLoginHooks[i] || PreLoginHooks[i]->bDeleteMe )
								{
									PreLoginHooks[i] = Found;
									debugf(NAME_XC_Engine, TEXT("Added %s to PreLoginHooks list slot %i"), Found->GetFullName(), i );
									break;
								}
						}
					}
				}
			}
		}
		Ar.Logf( TEXT("No actor name was input") );
	}
	else if ( ParseCommand(&Str,TEXT("AdminLoginHook")) )
	{
		TCHAR ActorName[65];
		if ( ParseToken( Str, ActorName, 65, 1 ) )
		{
			AActor* Found = FindNamedActor( GetActiveLevel(this), ActorName);
			if ( !Found )
				Ar.Logf( TEXT("Actor %s not found in level"), ActorName );
			else
			{
				UFunction* fAdminLoginHook = FindBaseFunction( Found->GetClass(), TEXT("AdminLoginHook") );
				if ( !fAdminLoginHook )
					Ar.Logf( TEXT("Class %s doesn't have a AdminLoginHook function"), Found->GetClass()->GetName() );
				else
				{
					if ( fAdminLoginHook->ParmsSize != 8 ) //Reference + BITFIELD
						Ar.Logf( TEXT("%s.AdminLoginHook has incorrect parameter block size: %i, should be 8"), ActorName, fAdminLoginHook->ParmsSize );
					else
					{
						UProperty* Parm = (UProperty*)fAdminLoginHook->Children;
						if ( !(Parm->PropertyFlags & CPF_OutParm) && Parm->GetClass() == UObjectProperty::StaticClass() && ((UObjectProperty*)Parm)->PropertyClass == APlayerPawn::StaticClass() )
						{
							Parm = (UProperty*)Parm->Next;
							if ( Parm->GetClass() == UBoolProperty::StaticClass() )
							{
								AdminLoginHook = Found;
								debugf(NAME_XC_Engine, TEXT("Set %s as AdminLoginHook"), Found->GetFullName() );
							}
							else
								Ar.Logf( TEXT("%s.AdminLoginHook has incorrect return type, should be BOOL"), ActorName );
						}
						else
							Ar.Logf( TEXT("%s.AdminLoginHook has incorrect parameter 1 type, should be PLAYERPAWN"), ActorName);
					}
				}
			}
		}
		Ar.Logf( TEXT("No actor name was input") );
	}
	else if ( XCGESystems.Exec( Cmd, Ar) )
		return 1;
	else
		return Super::Exec( Cmd, Ar);
	return 1;
	unguard;
}

void UXC_GameEngine::PerformSetCommand( const TCHAR* Str, FOutputDevice& Ar, UBOOL bNotifyObjectOfChange )
{
	// Set a class default variable, names are up to 64 chars and classname can go as (Package.Class)
	TCHAR ClassName[128], PropertyName[64], Number[128];
	UClass* Class;
	UProperty* Property;
	if ( ParseToken( Str, ClassName, ARRAY_COUNT(ClassName), 1 )
	&&	(Class=FindObject<UClass>( ANY_PACKAGE, ClassName))!=NULL )
	{
		if ( ParseToken( Str, PropertyName, ARRAY_COUNT(PropertyName), 1 )
		&&	(Property=FindField<UProperty>( Class, PropertyName))!=NULL )
		{
			while ( *Str==' ')
				Str++;
			if ( Property->ArrayDim == 1 )
			{
				UObject::GlobalSetProperty( Str, Class, Property, Property->Offset, bNotifyObjectOfChange );
				return;
			}
			const TCHAR* Tester = Str;
			if ( !(*Tester) ) //Nothing to do here
				return;
			while ( *Tester!=' ' && *Tester ) //Go ahead until i've gone thru all the token
				Tester++;
			while ( *Tester==' ')
				Tester++;

			if ( !(*Tester) ) //There was only one token, use Str directly as parameter
				UObject::GlobalSetProperty( Str, Class, Property, Property->Offset, bNotifyObjectOfChange );
			else if ( ParseToken( Str, Number, 64, 1) ) //Parse the parameter we just went through
			{
				while ( *Str==' ') //And remove spaces from main stream
					Str++;
				INT Idx = Clamp(appAtoi( Number), 0, Property->ArrayDim - 1);
				UObject::GlobalSetProperty( Str, Class, Property, Property->Offset + Idx * Property->ElementSize, bNotifyObjectOfChange );
			}
			else
				debugf(NAME_XC_Engine, TEXT("PerformSetCommand: SOMETHING IS NOT RIGHT!"));;
		}
		else
			Ar.Logf( NAME_ExecWarning, TEXT("Unrecognized property %s"), PropertyName );
	}
	else 
		Ar.Logf( NAME_ExecWarning, TEXT("Unrecognized class %s"), ClassName );
}

void UXC_GameEngine::Init()
{
	guard(UXC_GameEngine::Init);

	RegisterNames();

	debugf( NAME_Init, TEXT("========= XC_ENGINE - Test build %i ========="), Version);
	debugf( NAME_Init, TEXT("================= By Higor =================="));
	if ( b451Setup )
		debugf( NAME_Init, TEXT("Detected version 451 GameEngine"));

	//This probably fixes a v451 init crash
	(&GPendingLevel)[b451Setup] = NULL;

	StaticLoadClass( UXC_CoreStatics::StaticClass(), NULL, TEXT("XC_Core.XC_CoreStatics"), NULL, LOAD_NoWarn | LOAD_Quiet, NULL );
	
	GetDefault<UXC_CoreStatics>()->XC_Engine_Version = Version;
	XCGE_Defaults->CoreVersion = GetDefault<UXC_CoreStatics>()->XC_Core_Version;
	CoreVersion = XCGE_Defaults->CoreVersion;
	HookNatives( true);

	if ( UNetConnection::StaticClass()->GetPropertiesSize() == 16088 )
	{
		b440Net = true;
		XCGE_Defaults->b440Net = true;
		debugf( NAME_Init, TEXT("Detected version 440 net structures") );
	}

	//Swap UGameEngine and UXC_GameEngine property sizes temporarily to prevent an unwanted assertion
	Exchange( UGameEngine::StaticClass()->PropertiesSize, UXC_GameEngine::StaticClass()->PropertiesSize );
	UGameEngine::Init();

	FXC_TravelManager* TravelManager = new FXC_TravelManager;
	TravelManager->Engine = this;
	TravelManager->PostMapChange( GetActiveLevel(this) );
	debugf( NAME_Init, TEXT("XC_Engine Travel Manager initialized"));

	FXC_TimeManager* TimeManager = new FXC_TimeManager;
	debugf( NAME_Init, TEXT("XC_Engine Time Manager initialized"));

	XCGESystems.GenSystems.AddItem( TravelManager);
	XCGESystems.GenSystems.AddItem( TimeManager);
	XCGESystems.Init();
//	GSecondsPerCycle *= 2;
//	HookEngine( GetActiveLevel(this) );
	GetDefault<UXC_CoreStatics>()->XC_Engine_Version = Version;
	unguard;
}

//Graceful exit
void UXC_GameEngine::Exit()
{
	UnHookEngine();
	if ( CollisionCacus )
		delete CollisionCacus;
	CollisionCacus = NULL;
	XCGESystems.Exit();
	if ( XCGESystems.GenSystems.Num() )
	{
		for ( INT i=0 ; i<XCGESystems.GenSystems.Num() ; i++ )
		if ( XCGESystems.GenSystems(i) )
			delete XCGESystems.GenSystems(i);
		XCGESystems.GenSystems.Empty();
	}

	Super::Exit();
}

void UXC_GameEngine::Destroy()
{
	guard( UXC_GameEngine::Destroy() );
	debugf( TEXT("DESTROY XC_GAMEENGINE") );
	Super::Destroy();
	//Manual cleanup needed on v451 to avoid crash
	if ( b451Setup )
	{
		UGameEngine* GE = this + 1;
		SafeEmpty( GE->ServerActors );
		SafeEmpty( GE->ServerPackages );
		if ( GE->LastURL.Protocol.Len() )
			GE->LastURL.Protocol.Empty();
		if ( GE->LastURL.Host.Len() )
			GE->LastURL.Host.Empty();
		GE->LastURL.Port = 0;
		if ( GE->LastURL.Map.Len() )
			GE->LastURL.Map.Empty();
		SafeEmpty( GE->LastURL.Op);
		if ( GE->LastURL.Portal.Len() )
			GE->LastURL.Portal.Empty();
		appMemzero( &LastURL, sizeof(class FURL) + 28);
	}
	unguard;
}

void UXC_GameEngine::Tick( FLOAT DeltaSeconds)
{
	guard(UXC_GameEngine::Tick);

	//Force map redetection on next call, move to external?
	SafeEmpty( MapCache);
	LastMapIdx = 0;

	ULevelBase* GPendingBase = (ULevelBase*) *(&GEntry + 1 + b451Setup);
	//Reset download manager hook if pending level is gone
	if ( bHTTP_DLMGR && !GPendingBase )
		bHTTP_DLMGR = false;

	//GPendingLevel is up, attempt to find a valid redirect mgr
	if ( !bHTTP_DLMGR && GPendingBase && GPendingBase->NetDriver && GPendingBase->NetDriver->ServerConnection )
	{
		TArray<FDownloadInfo>* DI = (TArray<FDownloadInfo>*) (((DWORD) &GPendingBase->NetDriver->ServerConnection->DownloadInfo) + b440Net * 20);
		for ( INT i=0 ; i<DI->Num() ; i++ )
		{
			if ( (*DI)(i).ClassName == TEXT("XC_IpDrv.XC_HTTPDownload") )
			{
				bHTTP_DLMGR = true;
				break;
			}
			if ( (*DI)(i).ClassName == TEXT("IpDrv.HTTPDownload") )
			{
				UClass* DLC = StaticLoadClass( UXC_Download::StaticClass(), NULL, TEXT("XC_IpDrv.XC_HTTPDownload"), NULL, LOAD_NoWarn | LOAD_Quiet, NULL );
				if ( DLC )
				{
					UClass* OldDLC = (*DI)(i).Class;
					debugf( NAME_XC_Engine, TEXT("Replacing HTTP download manager with LZMA enhanced version.") );
					(*DI)(i).ClassName = TEXT("XC_IpDrv.XC_HTTPDownload");
					(*DI)(i).Class = DLC;
					UDownload** UDL = (UDownload**) (((DWORD) DI) - 4);
					if ( *UDL && (*UDL)->GetClass() == OldDLC )
					{
						debugf( NAME_XC_Engine, TEXT("Cancelling current download and restarting...") );
						GPendingBase->NetDriver->ServerConnection->ReceiveFile( (*UDL)->PackageIndex);
					}
				}
				bHTTP_DLMGR = true;
				break;
			}
		}
	}

	bCanEditPackageMap = false; //Disallow editing package map
	ULevel* TestLevel = GetActiveLevel( this);
	if ( TestLevel )
	{
		if ( bGCEnded )
		{
			if ( !Client && !bHackingTracker && TestLevel->BrushTracker )
				bHackingTracker = true;
			bGCEnded = false;
		}
	
		if ( bFasterUpload && TestLevel->NetDriver )
		{
			for ( INT i=0; i<TestLevel->NetDriver->ClientConnections.Num() ; i++ )
			{
				if ( TestLevel->NetDriver->ClientConnections(i) && !TestLevel->NetDriver->ClientConnections(i)->Actor )
				{
					if ( TestLevel->NetDriver->ClientConnections(i)->CurrentNetSpeed != 1000001 )
						TestLevel->NetDriver->ClientConnections(i)->ConfiguredInternetSpeed = TestLevel->NetDriver->ClientConnections(i)->CurrentNetSpeed;
					TestLevel->NetDriver->ClientConnections(i)->CurrentNetSpeed = 1000001;
				}
				else if ( TestLevel->NetDriver->ClientConnections(i) && TestLevel->NetDriver->ClientConnections(i)->Actor && 
							TestLevel->NetDriver->ClientConnections(i)->CurrentNetSpeed == 1000001 )
				{
					TestLevel->NetDriver->ClientConnections(i)->CurrentNetSpeed = TestLevel->NetDriver->ClientConnections(i)->ConfiguredInternetSpeed;
					if ( TestLevel->NetDriver->ClientConnections(i)->CurrentNetSpeed == 1000001 )
						TestLevel->NetDriver->ClientConnections(i)->CurrentNetSpeed = 20000;
				}
			}
		}
	}
	FLOAT NewDelta = DeltaSeconds;
	XCGESystems.Tick( DeltaSeconds);
	if ( !bDisableTimingFix )
	{
		FXC_TimeManager* TimeManager = (FXC_TimeManager*) XCGESystems.FindByType( TEXT("TimeManager") );
		if ( TimeManager )
			NewDelta *= TimeManager->Factor;
	}
	
	Super::Tick( NewDelta);
	
	unguard;
}

//For caching non-native Actor references
struct FClassPropertyCacheORef : public FClassPropertyCache
{
	FClassPropertyCacheORef( UClass* MasterClass=UObject::StaticClass() )
	:	FClassPropertyCache( MasterClass)
	{};
	FClassPropertyCacheORef( FClassPropertyCache* InNext, UClass* InClass)
	:	FClassPropertyCache( InNext, InClass)
	{};

	UBOOL AcceptProperty( UProperty* Property)
	{
		if ( Property->PropertyFlags & CPF_Native )
			return 0;
		UObjectProperty* ObjProp = Cast<UObjectProperty>(Property);
		return ObjProp && ObjProp->PropertyClass->IsChildOf(AActor::StaticClass());
	}
	FClassPropertyCache* CreateParent( FMemStack& Mem)
	{
		return new(Mem) FClassPropertyCacheORef( Next, Class->GetSuperClass() );
	}
};

void UXC_GameEngine::CleanupTransientRefs()
{
	guard( UXC_GameEngine::CleanupTransientRefs);

	ULevel* EntryLevel = GetEntryLevel(this);
	ULevel* ActiveLevel = GetActiveLevel(this);

	if ( !ActiveLevel || (ActiveLevel == EntryLevel) )
		return;

	FMemMark Mark(GMem);
	FClassPropertyCacheORef* ObjCache = new(GMem) FClassPropertyCacheORef();
	ObjCache->GrabProperties( GMem);

	ObjCache = new(GMem) FClassPropertyCacheORef( ObjCache, UScriptedTexture::StaticClass() );
	ObjCache->GrabProperties( GMem);
	FClassPropertyCacheORef* ScriptedTextureCache = ObjCache;
	
	UBOOL bFirstReference = true;
	guard( Scanner);
	for ( TObjectIterator<UObject> It; It; ++It )
	{
		if ( It->GetOuter() )
		{
			if ( It->IsA(UPlayer::StaticClass()) || It->IsA(USubsystem::StaticClass()) )
				continue;
			FClassPropertyCache* CurCache = NULL;
			guard( EvaluateType);
			if ( It->IsA(UScriptedTexture::StaticClass()) )
				CurCache = ScriptedTextureCache;
			else
			{
				UObject* TheOuter = It->GetOuter();
				while ( TheOuter && (TheOuter->GetOuter() != TheOuter) )
				{
					if ( (TheOuter == GetTransientPackage()) || (EntryLevel && (TheOuter == EntryLevel)) )
					{
						CurCache = ObjCache->GetCache( It->GetClass() );
						if ( !CurCache )
						{
							ObjCache = new(GMem) FClassPropertyCacheORef( ObjCache, It->GetClass() );
							ObjCache->GrabProperties(GMem);
							CurCache = ObjCache;
						}
						break;
					}
					TheOuter = TheOuter->GetOuter();
				}
			}
			unguardf( ( TEXT("(%s)"), It->GetFullName() ) );
//			if ( ClassMap(Idx).TotalProps <= 0 )
//				continue;
//			if ( Idx > 0 )
//				debugf( NAME_Log, TEXT("[XCGE]--- Scanning object %s"), It->GetName() );
			guard( Clearing);
			for ( ; CurCache ; CurCache=CurCache->Parent )
			{
				for ( FPropertyCache* PropCache=CurCache->Properties ; PropCache ; PropCache=PropCache->Next )
				{
					UProperty* Prop = PropCache->Property;
//					debugf( NAME_Log, TEXT("[XCGE]---- Scanning property %s"), Prop->GetName() );
					check( Prop);
					UObject* TheObj = *It;
					AActor** Ref = (AActor**) ( DWORD (TheObj) + Prop->Offset);
					INT j = 0;
					guard( Array);
					for ( ; j<Prop->ArrayDim ; j++ )
					{
						if ( *Ref && ((*Ref)->GetOuter() == ActiveLevel->GetOuter() ) )
						{
							if ( bEnableDebugLogs )
							{
								guard( DebugLog);
								if ( bFirstReference )
								{
									bFirstReference = false;
									debugf( NAME_XC_Engine, TEXT("[WARNING] Found dangerous actor references pointing to Level actors, nulling out:"));
								}
								debugf( NAME_XC_Engine, TEXT("[WARNING]=== %s: %s[%i]=%s"), It->GetFullName(), Prop->GetName(), j, (*Ref)->GetFullName() );
								unguard;
							}
							*Ref = NULL;
						}
						Ref++;
					}
					unguardf(( TEXT("(%i/%i)"), j, Prop->ArrayDim ));
				}
//				Idx = ClassMap(Idx).ParentIdx;
			}
			unguardf( ( TEXT("(%s)"), It->GetFullName() ) );
		}
	}
	unguard;
	if ( !bFirstReference )
		debugf( NAME_XC_Engine, TEXT("[WARNING] Transient cleaner has removed all dangerous references"));
	Mark.Pop();
	unguard;
}

//==============================================================================
// FCollisionCacus experiment
//==============================================================================

UBOOL UXC_GameEngine::SetupCollisionCacus( ULevel* Lev)
{
	guard(UXC_GameEngine::SetupCollisionCacus);
	check( Lev);
	if ( CollisionCacus && (Lev->Hash == CollisionCacus) ) //Never override existing grid
		return 0;
	if ( appStricmp( Lev->GetOuter()->GetName(), TEXT("Entry")) == 0 )
		return 0;
	if ( !CollisionCacus )
	{
		if ( Lev->Hash )
			Lev->SetActorCollision(0);
		debugf( NAME_XC_Engine, TEXT("Setting up collision octree [Engine Hook] for %s"), Lev->GetOuter() ? Lev->GetOuter()->GetName() : TEXT("[Unspecified]") );
		CollisionCacus = new FCollisionCacus(Lev);
		guard( AddActors);
		for( INT i=0 ; i<Lev->Actors.Num() ; i++ )
			if( Lev->Actors(i) && !Lev->Actors(i)->bDeleteMe && Lev->Actors(i)->bCollideActors )
				CollisionCacus->AddActor( Lev->Actors(i) );
		unguard;
	}
	Lev->Hash = CollisionCacus;
	return 1;
	unguard;
}


//==============================================================================
// Connection handler
//==============================================================================
UBOOL AXC_ConnectionHandler::Tick( FLOAT DeltaTime, enum ELevelTick TickType )
{
	if ( XLevel->NetDriver )
	{
		INT i;
		for ( i=0 ; i<ExtraTCPQueries ; i++ )
			XLevel->NetDriver->TickDispatch( 0.f);

		INT EConnCount = 0;
		for ( i=0 ; i<XLevel->NetDriver->ClientConnections.Num() ; i++ )
			if ( XLevel->NetDriver->ClientConnections(i) && !XLevel->NetDriver->ClientConnections(i)->OpenChannels.Num() )
				EConnCount++;
		FLOAT TimeOut = EConnCount >= CriticalConnCount ? CriticalTimeout : DatalessTimeout;
		for ( i=0 ; i<XLevel->NetDriver->ClientConnections.Num() ; i++ )
		{
			UNetConnection* Conn = XLevel->NetDriver->ClientConnections(i);
			if ( Conn && !Conn->OpenChannels.Num() && (XLevel->NetDriver->Time - Conn->LastReceiveTime > TimeOut) )
			{
				debugf( NAME_XC_Engine, TEXT("Empty connection after %f seconds."), XLevel->NetDriver->Time - Conn->LastReceiveTime );
				Conn->State = USOCK_Closed;
			}
		}
	}
	return Super::Tick( DeltaTime, TickType);
}


//==============================================================================
// Dummy Pawn functions and hooks
//==============================================================================

#ifndef DISABLE_SCRIPT

//The actual context is a AActor
void AXC_DummyPawn::NewMapName(FFrame &Stack, RESULT_DECL)
{
	guard( AXC_DummyPawn::NewMapName );
	P_GET_STR( NameEnding);
	P_GET_STR( MapName);
	P_GET_INT( Dir);
	P_FINISH;

	UXC_GameEngine* XC = (UXC_GameEngine*) XLevel->Engine;
	if ( !XC->MapCache.Num() || (NameEnding != XC->MapCachedPrefix) )
	{
		XC->MapCachedPrefix = NameEnding;
		if ( XC->bEnableDebugLogs )
			debugf( NAME_XC_Engine, TEXT("MapCache needs reloading...") );

		for ( int iP=0 ; iP<GSys->Paths.Num() ; iP++ )
		{
			if ( appStricmp( *(GSys->Paths(iP).Right(5)), TEXT("*.unr") ) == 0 )
			{
				FString RootDir = GSys->Paths(iP).Left( GSys->Paths(iP).Len() - 5);
				TArray<FString> Files = GFileManager->FindFiles( *(RootDir + NameEnding + TEXT("*.unr") ), true, false);
				if ( XC->bSortMaplistByFolder && !XC->bSortMaplistGlobal )
					SortStringsA( &Files );
				if ( XC->bEnableDebugLogs )
					debugf( NAME_XC_Engine, TEXT("Searching through directory %s => Found %i results"), *RootDir, Files.Num() );
				int OldSize = XC->MapCache.Num();

				//Optimal array merging
				XC->MapCache.Add( Files.Num() );
				appMemcpy_amd( &XC->MapCache(OldSize), &Files(0), Files.Num() * sizeof(class FString) );
				appFree( Files.GetData() );
				appMemzero( &Files, sizeof(class TArray<FString>) );

//				XC->MapCache.AddZeroed( Files.Num() );
//				for ( int iM = 0 ; iM < Files.Num() ; iM++ )
//					XC->MapCache( OldSize + iM) = Files(iM);
			}
		}
		if ( XC->bSortMaplistGlobal )
			SortStringsA( &(XC->MapCache) );

	}
	if ( XC->MapCache.Num() > 0 )
	{
		if ( !MapName.Len() )
			XC->LastMapIdx = 0;
		else if ( XC->MapCache(XC->LastMapIdx) == MapName ) //Chained search
			XC->LastMapIdx += Dir;
		else
		{
			for ( int i = 0 ; i<XC->MapCache.Num() ; i++ )
				if ( XC->MapCache(i) == MapName )
				{
					XC->LastMapIdx = i + Dir;
					break;
				}
		}
	
		while ( XC->LastMapIdx >= XC->MapCache.Num() )
			XC->LastMapIdx -= XC->MapCache.Num();
		while ( XC->LastMapIdx < 0 )
			XC->LastMapIdx += XC->MapCache.Num();
		*(FString*)Result = XC->MapCache( XC->LastMapIdx );
	}
	unguard;
}


static UBOOL PCanSee( AActor* Seen, APlayerPawn* Other)
{
	guard( XC_Engine::PCanSee);
	check( Other && Seen);
	FVector aVec = Seen->Location - Other->Location;
	if ( aVec.IsZero() )
		return 0;
	if ( aVec.SizeSquared() > 2560000 )
		return 0;
	if ( aVec.SizeSquared() > (aVec + Other->ViewRotation.Vector()).SizeSquared() )
		return 0;
	if ( !aVec.Normalize() )
		return 0;
	FLOAT aFov = Other->FovAngle / 114.6; //Real angle outside of the center
	FLOAT aDot = aVec|Other->ViewRotation.Vector(); //This is the cosine of our angle
	if ( aDot < appCos(aFov * 1.2) )
		return 0;
	FCheckResult Hit;
	return Seen->XLevel->SingleLineCheck( Hit, Seen, Seen->Location, Other->Location + FVector(0,0,Other->BaseEyeHeight), TRACE_Level, FVector(0,0,0), 0|NF_NotVisBlocking);
	unguard;
}

//Called mostly on Actor context
void AXC_DummyPawn::NewPlayerCanSeeMe(FFrame &Stack, RESULT_DECL)
{
	guard( XC_Engine::NewPlayerCanSeeMe);
	P_FINISH;

	*(UBOOL*)Result = 0;
	if ( XLevel )
	{
		if ( XLevel->Engine->Client && XLevel->Engine->Client->Viewports.Num() && XLevel->Engine->Client->Viewports(0) && XLevel->Engine->Client->Viewports(0)->Actor ) //Local player
			if ( PCanSee( this, XLevel->Engine->Client->Viewports(0)->Actor) )
			{
				*(UBOOL*)Result = 1;
				return;
			}
		if ( XLevel->NetDriver && XLevel->NetDriver->ClientConnections.Num() ) //Clients
		{
			for ( INT ip=0 ; ip < XLevel->NetDriver->ClientConnections.Num() ; ip++ )
				if ( XLevel->NetDriver->ClientConnections(ip) && XLevel->NetDriver->ClientConnections(ip)->Actor )
					if ( PCanSee( this, XLevel->NetDriver->ClientConnections(ip)->Actor) )
					{
						*(UBOOL*)Result = 1;
						return;
					}
		}
	}

	unguard;
}


void AXC_DummyPawn::NewDynArrayElement( FFrame& Stack, RESULT_DECL )
{
	// Grab the index
	INT Index=0;
	Stack.Step( Stack.Object, &Index );

	// Grab the property
	GProperty = NULL;
	Stack.Step( this, NULL );

	if( GProperty && GPropAddr )
	{
		FArray* Array=(FArray*)GPropAddr;
		UArrayProperty* ArrayProp = (UArrayProperty*)GProperty;
		if( Index>=Array->Num() || Index<0 )
		{
			//if we are returning a value, check for out-of-bounds
			if ( Result || Index<0 )
			{
				Stack.Logf( NAME_Error, TEXT("Accessed array '%s' out of bounds (%i/%i)"), ArrayProp->GetName(), Index, Array->Num() );
				GPropAddr = 0;
				if (Result)
					appMemzero( Result, ArrayProp->Inner->ElementSize );
				return;
			}
			//if we are setting a value, allow the array to be resized
			else
				Array->AddZeroed(ArrayProp->Inner->ElementSize,Index-Array->Num()+1);
		}
		GPropAddr = (BYTE*)Array->GetData() + Index * ArrayProp->Inner->ElementSize;

		// Add scaled offset to base pointer.
		if( Result )
			ArrayProp->Inner->CopySingleValue( Result, GPropAddr );
	}
}

void AXC_DummyPawn::NewDynArrayLength( FFrame& Stack, RESULT_DECL )
{
	GProperty = NULL;
	Stack.Step( this, NULL );
	UBOOL bSetSize = false;
	INT NewSize = 0;
	if ( *(Stack.Code) != EX_EndFunctionParms )
	{
		Stack.Step( Stack.Object, &NewSize);
		if ( GProperty->PropertyFlags & CPF_Const )
			debugf( NAME_ScriptWarning, TEXT("Set length of Dynamic Array '%s' marked as 'const'"), GProperty->GetName() );
		else if ( NewSize < 0 )
			debugf( NAME_ScriptWarning, TEXT("Set length of Dynamic Array '%s' to a negative size: %i"), GProperty->GetName(), NewSize );
		else
			bSetSize = true;
	}
	P_FINISH;

	if (GPropAddr && GProperty)
	{
		FArray* Array=(FArray*)GPropAddr;
		UArrayProperty* ArrayProp = (UArrayProperty*)GProperty;
		if ( bSetSize )
		{
			if ( NewSize > Array->Num() )
				Array->AddZeroed(ArrayProp->Inner->ElementSize, NewSize-Array->Num());
			else if (NewSize < Array->Num())
			{
				for (INT i=Array->Num()-1; i>=NewSize; i--)
					ArrayProp->Inner->DestroyValue((BYTE*)Array->GetData() + ArrayProp->Inner->ElementSize*i);
				Array->Remove(NewSize, Array->Num()-NewSize, ArrayProp->Inner->ElementSize );
			}
		}
		*(INT*)Result = Array->Num();
	}
}

void AXC_DummyPawn::NewDynArrayInsert( FFrame& Stack, RESULT_DECL )
{
	GProperty = NULL;
	Stack.Step( this, NULL );
	UArrayProperty* ArrayProperty = Cast<UArrayProperty>(GProperty);
	FArray* Array=(FArray*)GPropAddr;

	P_GET_INT(Index);
	P_GET_INT_OPTX(Count,1);
	P_FINISH;

	if ( ArrayProperty->PropertyFlags & CPF_Const )
		debugf( NAME_ScriptWarning, TEXT("Insert in Dynamic Array '%s' marked as 'const'"), ArrayProperty->GetName() );
	else if (Array && Count)
	{
		if ( Count < 0 )
		{
			Stack.Logf( TEXT("Attempt to insert a negative number of elements '%s'"), ArrayProperty->GetName() );
			return;
		}
		if ( Index < 0 || Index > Array->Num() )
		{
			Stack.Logf( TEXT("Attempt to insert %i elements at %i an %i-element array '%s'"), Count, Index, Array->Num(), ArrayProperty->GetName() );
			Index = Clamp(Index, 0,Array->Num());
		}
		Array->InsertZeroed( Index, Count, ArrayProperty->Inner->ElementSize);
		*(UBOOL*)Result = true;
	}
}

void AXC_DummyPawn::NewDynArrayRemove( FFrame& Stack, RESULT_DECL )
{	
	GProperty = NULL;
	Stack.Step( this, NULL );
	UArrayProperty* ArrayProperty = Cast<UArrayProperty>(GProperty);
	FArray* Array=(FArray*)GPropAddr;

	P_GET_INT(Index);
	P_GET_INT_OPTX(Count, 1);
	P_FINISH;
	if ( ArrayProperty->PropertyFlags & CPF_Const )
		debugf( NAME_ScriptWarning, TEXT("Remove in Dynamic Array '%s' marked as 'const'"), ArrayProperty->GetName() );
	else if (Array && Count)
	{
		if ( Count < 0 )
		{
			Stack.Logf( TEXT("Attempt to remove a negative number of elements '%s'"), ArrayProperty->GetName() );
			return;
		}
		if ( Index < 0 || Index >= Array->Num() || Index + Count > Array->Num() )
		{
			if (Count == 1)
				Stack.Logf( TEXT("Attempt to remove element %i in an %i-element array '%s'"), Index, Array->Num(), ArrayProperty->GetName() );
			else
				Stack.Logf( TEXT("Attempt to remove elements %i through %i in an %i-element array '%s'"), Index, Index+Count-1, Array->Num(), ArrayProperty->GetName() );
			Index = Clamp(Index, 0,Array->Num());
			if ( Index + Count > Array->Num() )
				Count = Array->Num() - Index;
		}

		for (INT i=Index+Count-1; i>=Index; i--)
			ArrayProperty->Inner->DestroyValue((BYTE*)Array->GetData() + ArrayProperty->Inner->ElementSize*i);
		Array->Remove( Index, Count, ArrayProperty->Inner->ElementSize);
		*(UBOOL*)Result = true;
	}
}

void AXC_DummyPawn::execCollidingActors( FFrame& Stack, RESULT_DECL )
{
	P_GET_OBJECT(UClass,BaseClass);
	P_GET_ACTOR_REF(OutActor);
	P_GET_FLOAT(Radius);
	P_GET_VECTOR_OPTX(TraceLocation,Location);
	P_FINISH;

	BaseClass = BaseClass ? BaseClass : AActor::StaticClass();
	FMemMark Mark(GMem);
	FIteratorActorList* Link=GetLevel()->Hash->ActorRadiusCheck( GMem, TraceLocation, Radius, 0 );

	PRE_ITERATOR;
		*OutActor = NULL;
		if ( Link )
		{
			while ( Link )
			{	//Actors can be de-collided or killed during the iterator, perform (almost) all checks!
				if( !Link->Actor || Link->Actor->bDeleteMe || !Link->Actor->bCollideActors || !Link->Actor->IsA(BaseClass) )
					Link=Link->GetNext();
				else
					break;
			}
			if ( Link )
			{
				*OutActor = Link->Actor;
				Link=Link->GetNext();
			}
		}
		if ( *OutActor == NULL ) 
		{
			Stack.Code = &Stack.Node->Script(wEndOffset + 1);
			break;
		}
	POST_ITERATOR;
	Mark.Pop();
}

void AXC_DummyPawn::NewLocalVariable( FFrame& Stack, RESULT_DECL )
{
	checkSlow(Stack.Object==this);
	checkSlow(Stack.Locals!=NULL);
	GProperty = (UProperty*)Stack.ReadObject();
	GPropAddr = Stack.Locals + GProperty->Offset;
	GPropActor = NULL;
	if( Result )
		GProperty->CopyCompleteValue( Result, GPropAddr );
}
void AXC_DummyPawn::NewInstanceVariable( FFrame& Stack, RESULT_DECL)
{
	GProperty = (UProperty*)Stack.ReadObject();
	GPropAddr = (BYTE*)this + GProperty->Offset;
	GPropActor = Cast<AActor>(this);
	if( Result )
		GProperty->CopyCompleteValue( Result, GPropAddr );
}
void AXC_DummyPawn::NewDefaultVariable( FFrame& Stack, RESULT_DECL )
{
	GProperty = (UProperty*)Stack.ReadObject();
	GPropAddr = &GetClass()->Defaults(GProperty->Offset);
	GPropActor = NULL;
	if( Result )
		GProperty->CopyCompleteValue( Result, GPropAddr );
}
void AXC_DummyPawn::NewVirtualFunction( FFrame& Stack, RESULT_DECL )
{
	FName FuncName = Stack.ReadName();
	guard(ScriptDebugV);
	CallFunction( Stack, Result, FindFunctionChecked(FuncName) );

	//HACK, THIS SHOULD GO IN UOBJECT::CALLFUNCTION
	if ( GPropActor && GProperty && (GProperty->PropertyFlags & CPF_Net) )
		GPropActor->bNetDirty = true;

	unguardf( (TEXT("%s.%s"), GetName(), *FuncName ) );
}

void AXC_DummyPawn::NewLet( FFrame& Stack, RESULT_DECL )
{
	// Get variable address.
	GPropAddr = NULL;
	Stack.Step( Stack.Object, NULL ); // Evaluate variable.
	if( !GPropAddr )
	{
		Stack.Logf( NAME_ScriptWarning, TEXT("Attempt to assign variable through None") );
		static BYTE Crud[1024];//@temp
		GPropAddr = Crud;
		appMemzero( GPropAddr, sizeof(FString) );
	}
	else if ( GPropActor && GProperty && (GProperty->PropertyFlags & CPF_Net) )
		GPropActor->bNetDirty = true; //FIXME - use object property instead for performance
	Stack.Step( Stack.Object, GPropAddr ); // Evaluate expression into variable.
}

void AXC_DummyPawn::NewLetBool( FFrame& Stack, RESULT_DECL )
{
	// Get variable address.
	GPropAddr = NULL;
	GProperty = NULL;
	GPropActor = NULL;
	Stack.Step( Stack.Object, NULL ); // Variable.
	BITFIELD*      BoolAddr     = (BITFIELD*)GPropAddr;
	UBoolProperty* BoolProperty = (UBoolProperty*)GProperty;
	INT Value=0;  // must be INT...execTrue/execFalse return 32 bits. --ryan.
	if ( GPropActor && GProperty && (GProperty->PropertyFlags & CPF_Net) )
		GPropActor->bNetDirty = true;
	Stack.Step( Stack.Object, &Value );
	if( BoolAddr )
	{
		checkSlow(BoolProperty->IsA(UBoolProperty::StaticClass()));
		if( Value ) *BoolAddr |=  BoolProperty->BitMask;
		else        *BoolAddr &= ~BoolProperty->BitMask;
	}
}

void AXC_DummyPawn::NewContext( FFrame& Stack, RESULT_DECL )
{
	// Get actor variable.
	UObject* NewContext=NULL;
	Stack.Step( this, &NewContext );

	// Execute or skip the following expression in the actor's context.
	if( NewContext != NULL )
	{
		Stack.Code += 3;
		Stack.Step( NewContext, Result );
	}
	else
	{
		if ( GProperty )
			Stack.Logf( NAME_Warning, TEXT("Accessed None '%s'"), GProperty->GetName() );
		else
			Stack.Logf( NAME_Warning, TEXT("Accessed None"));
		INT wSkip = Stack.ReadWord();
		BYTE bSize = *Stack.Code++;
		Stack.Code += wSkip;
		GPropAddr = NULL;
		GProperty = NULL;
		GPropActor = NULL;
		if( Result )
			appMemzero( Result, bSize );
	}
}

void AXC_DummyPawn::NewClassContext( FFrame& Stack, RESULT_DECL )
{
	// Get class expression.
	UClass* ClassContext=NULL;
	Stack.Step( Stack.Object, &ClassContext );

	// Execute expression in class context.
	if( ClassContext )
	{
		Stack.Code += 3;
		Stack.Step( ClassContext->GetDefaultObject(), Result );
	}
	else
	{
		if ( GProperty )
			Stack.Logf( NAME_Error, TEXT("Accessed null class context '%s'"), GProperty->GetName() );
		else
			Stack.Logf( NAME_Error, TEXT("Accessed null class context") );
		INT wSkip = Stack.ReadWord();
		BYTE bSize = *Stack.Code++;
		Stack.Code += wSkip;
		GPropAddr = NULL;
		GPropActor = NULL;
		GProperty = NULL;
		if( Result )
			appMemzero( Result, bSize );
	}
}

void AXC_DummyPawn::NewStructMember( FFrame& Stack, RESULT_DECL )
{
	// Get structure element.
	UProperty* Property = (UProperty*)Stack.ReadObject();

	// Get struct expression.
	UStruct* Struct = CastChecked<UStruct>(Property->GetOuter());
	BYTE* Buffer = (BYTE*)appAlloca(Struct->PropertiesSize);
	appMemzero( Buffer, Struct->PropertiesSize );
	GPropAddr = NULL;
	Stack.Step( this, Buffer );

	// Set result.
	GProperty = Property;
	GPropActor = Cast<AActor>(this);
	if( GPropAddr )
		GPropAddr += Property->Offset;
	if( Result )
		Property->CopyCompleteValue( Result, Buffer+Property->Offset );
	for( UProperty* P=Struct->ConstructorLink; P; P=P->ConstructorLinkNext )
		P->DestroyValue( Buffer + P->Offset );
}


void AXC_DummyPawn::execAddToPackagesMap( FFrame& Stack, RESULT_DECL )
{
	guard(AXC_DummyPawn::execAddToPackagesMap);
	ULinkerLoad* ToAdd = NULL;
	UBOOL bSelfLinker = 1;
	*(UBOOL*)Result = 0;
	UXC_GameEngine* XC = (UXC_GameEngine*) XLevel->Engine;
	if ( !XLevel->NetDriver || !XLevel->NetDriver->MasterMap )
		XC->bCanEditPackageMap = false;
	if ( *Stack.Code != EX_EndFunctionParms ) //Someone specified a package name
	{
		FString PackageName(0);
		Stack.Step( Stack.Object, &PackageName);
		bSelfLinker = 0;
		if ( XC->bCanEditPackageMap )
		{
			BeginLoad();
			ToAdd = GetPackageLinker( NULL, *PackageName, 0, NULL, NULL );
			EndLoad();
			if ( !ToAdd )
				debugf( NAME_Warning, TEXT("AddToPackagesMap: Linker not found for %s"), *PackageName );
		}
	}
	P_FINISH;
	if ( !XC->bCanEditPackageMap )
	{
		debugf( NAME_Warning, TEXT("AddToPackagesMap: Cannot edit the Master Map"));
		return;
	}

	if ( bSelfLinker )
	{
		ToAdd = GetClass()->GetLinker();
		if ( !ToAdd )
			debugf( NAME_Warning, TEXT("AddToPackagesMap: Actor %s has no linker!"), GetName() );
	}
	if ( !ToAdd )
		return;
	INT OldNum = XLevel->NetDriver->MasterMap->List.Num();
	XLevel->NetDriver->MasterMap->AddLinker( ToAdd );
	if ( OldNum != XLevel->NetDriver->MasterMap->List.Num() ) //Linker was inserted, Compute
	{
		XLevel->NetDriver->MasterMap->Compute();
		*(UBOOL*)Result = 1;
	}
	unguard;
}

void AXC_DummyPawn::execIsInPackageMap( FFrame& Stack, RESULT_DECL)
{
	guard(AXC_DummyPawn::execAddToPackagesMap);
	FString PackageName(0);
	ULinkerLoad* ToCheck = NULL;
	Stack.Step( Stack.Object, &PackageName);
	P_GET_UBOOL_OPTX( bServerPackagesOnly, 0);
	P_FINISH;
	*(UBOOL*)Result = 0;

	if ( !XLevel->NetDriver || !XLevel->NetDriver->MasterMap ) //Not a net server
		return;

	if ( !PackageName.Len() ) //Package this actor was created from
	{
		PackageName = GetClass()->GetName();
		ToCheck = GetClass()->GetLinker();
	}

	if ( bServerPackagesOnly )
	{
		UGameEngine* GameEngine = Cast<UGameEngine>(XLevel->Engine);
		check( GameEngine);
		TArray<FString>* SP = (TArray<FString>*)  (((DWORD) &GameEngine->ServerPackages) + XCGE_Defaults->b451Setup * 4 );
		for ( INT i=0 ; i<SP->Num() ; i++ )
			if ( (*SP)(i) == PackageName )
			{
				*(UBOOL*)Result = 1;
				break;
			}
	}
	else
	{
		if ( !ToCheck )
		{
			BeginLoad();
			ToCheck = GetPackageLinker( NULL, *PackageName, 0, NULL, NULL );
			EndLoad();
			if ( !ToCheck )
			{
				debugf( NAME_Warning, TEXT("IsInPackageMap: Linker not found for %s"), *PackageName );
				return;
			}
		}
		TArray<FPackageInfo>* PList = &XLevel->NetDriver->MasterMap->List;
		for ( INT i=0 ; i<PList->Num() ; i++ )
			if ( (*PList)(i).Linker == ToCheck )
			{
				*(UBOOL*)Result = 1;
				break;
			}
	}
	unguard;
}


#endif
